#ifndef CONSTANTS_H_
#define CONSTANTS_H_

#include <string>

namespace Constants {
    extern const double EPS;
    extern const double INF;

    extern const std::string default_utility_name;
    extern const std::string default_counterfactual_name;
    extern const std::string default_counterfactual_ev_name;
    extern const std::string default_reach_prob_name;
    extern const std::string default_opponent_reach_prob_name;
    extern const std::string default_action_set_size_name;
    extern const std::string default_subtree_size_name;
    extern const std::string default_chance_prob_name;
    extern const std::string default_auxiliary_result_prefix;
};

#endif