#include "Basic/Constants.h"

const double Constants::EPS = 1e-9;
const double Constants::INF = 1e9;

const std::string Constants::default_utility_name = "utility";
const std::string Constants::default_counterfactual_name = "counterfactual";
const std::string Constants::default_counterfactual_ev_name = "counterfactual_ev";
const std::string Constants::default_reach_prob_name = "reach_prob";
const std::string Constants::default_opponent_reach_prob_name = "opponent_reach_prob";
const std::string Constants::default_action_set_size_name = "action_set_size";
const std::string Constants::default_subtree_size_name = "subtree_size";
const std::string Constants::default_chance_prob_name = "chance_prob";
const std::string Constants::default_auxiliary_result_prefix = "auxiliary_result_";